package javax.swing.text;

import java.awt.Rectangle;
import java.awt.Graphics;
import java.awt.Shape;
import javax.swing.event.DocumentEvent;
import java.awt.Container;
import java.util.ArrayList;
import java.util.List;

public class AsyncBoxView extends View
{
    int axis;
    List<AsyncBoxView.ChildState> stats;
    float majorSpan;
    boolean estimatedMajorSpan;
    float minorSpan;
    protected AsyncBoxView.ChildLocator locator;
    float topInset;
    float bottomInset;
    float leftInset;
    float rightInset;
    AsyncBoxView.ChildState minRequest;
    AsyncBoxView.ChildState prefRequest;
    boolean majorChanged;
    boolean minorChanged;
    Runnable flushTask;
    AsyncBoxView.ChildState changing;
    
    public AsyncBoxView(Element element, int axis) {
        super(element);
        this.stats = new ArrayList();
        this.axis = axis;
        this.locator = new AsyncBoxView.ChildLocator();
        this.flushTask = new AsyncBoxView.FlushTask();
        this.minorSpan = 32767.0f;
        this.estimatedMajorSpan = false;
    }
    
    public int getMajorAxis() {
        return this.axis;
    }
    
    public int getMinorAxis() {
        return this.axis == 0?1:0;
    }
    
    public float getTopInset() {
        return this.topInset;
    }
    
    public void setTopInset(float topInset) {
        this.topInset = topInset;
    }
    
    public float getBottomInset() {
        return this.bottomInset;
    }
    
    public void setBottomInset(float bottomInset) {
        this.bottomInset = bottomInset;
    }
    
    public float getLeftInset() {
        return this.leftInset;
    }
    
    public void setLeftInset(float leftInset) {
        this.leftInset = leftInset;
    }
    
    public float getRightInset() {
        return this.rightInset;
    }
    
    public void setRightInset(float rightInset) {
        this.rightInset = rightInset;
    }
    
    protected float getInsetSpan(int n) {
        return (n == 0) ? (this.getLeftInset() + this.getRightInset()) : (this.getTopInset() + this.getBottomInset());
    }
    
    protected void setEstimatedMajorSpan(boolean estimatedMajorSpan) {
        this.estimatedMajorSpan = estimatedMajorSpan;
    }
    
    protected boolean getEstimatedMajorSpan() {
        return this.estimatedMajorSpan;
    }
    
    protected AsyncBoxView.ChildState getChildState(int n) {
        synchronized (this.stats) {
            if (n >= 0 && n < this.stats.size()) {
                return (AsyncBoxView.ChildState)this.stats.get(n);
            }
            return null;
        }
    }
    
    protected LayoutQueue getLayoutQueue() {
        return LayoutQueue.getDefaultQueue();
    }
    
    protected AsyncBoxView.ChildState createChildState(View view) {
        return new AsyncBoxView.ChildState(view);
    }
    
    protected synchronized void majorRequirementChange(AsyncBoxView.ChildState childState, float n) {
        if (!this.estimatedMajorSpan) {
            this.majorSpan += n;
        }
        this.majorChanged = true;
    }
    
    protected synchronized void minorRequirementChange(AsyncBoxView.ChildState childState) {
        this.minorChanged = true;
    }
    
    protected void flushRequirementChanges() {
        AbstractDocument abstractDocument = (AbstractDocument)this.getDocument();
        try {
            abstractDocument.readLock();
            View parent = null;
            boolean n = false;
            boolean n2 = false;
            synchronized (this) {
                synchronized (this.stats) {
                    int viewCount = this.getViewCount();
                    if (viewCount > 0 && (this.minorChanged || this.estimatedMajorSpan)) {
                        this.getLayoutQueue();
                        AsyncBoxView.ChildState childState = this.getChildState(0);
                        AsyncBoxView.ChildState childState2 = this.getChildState(0);
                        float majorSpan = 0.0f;
                        for (int i = 1; i < viewCount; ++i) {
                            AsyncBoxView.ChildState childState3 = this.getChildState(i);
                            if (this.minorChanged) {
                                if (childState3.min > childState.min) {
                                    childState = childState3;
                                }
                                if (childState3.pref > childState2.pref) {
                                    childState2 = childState3;
                                }
                            }
                            if (this.estimatedMajorSpan) {
                                majorSpan += childState3.getMajorSpan();
                            }
                        }
                        if (this.minorChanged) {
                            this.minRequest = childState;
                            this.prefRequest = childState2;
                        }
                        if (this.estimatedMajorSpan) {
                            this.majorSpan = majorSpan;
                            this.estimatedMajorSpan = false;
                            this.majorChanged = true;
                        }
                    }
                }
                if (this.majorChanged || this.minorChanged) {
                    parent = this.getParent();
                    if (parent != null) {
                        if (this.axis == 0) {
                            n = this.majorChanged;
                            n2 = this.minorChanged;
                        }
                        else {
                            n2 = this.majorChanged;
                            n = this.minorChanged;
                        }
                    }
                    this.majorChanged = false;
                    this.minorChanged = false;
                }
            }
            if (parent != null) {
                parent.preferenceChanged(this, (boolean)n, (boolean)n2);
                Container container = this.getContainer();
                if (container != null) {
                    container.repaint();
                }
            }
        }
        finally {
            abstractDocument.readUnlock();
        }
    }
    
    @Override
    public void replace(int n, int n2, View[] array) {
        synchronized (this.stats) {
            for (int i = 0; i < n2; ++i) {
                AsyncBoxView.ChildState childState = (AsyncBoxView.ChildState)this.stats.remove(n);
                float majorSpan = childState.getMajorSpan();
                childState.getChildView().setParent(null);
                if (majorSpan != 0.0f) {
                    this.majorRequirementChange(childState, -majorSpan);
                }
            }
            LayoutQueue layoutQueue = this.getLayoutQueue();
            if (array != null) {
                for (int j = 0; j < array.length; ++j) {
                    AsyncBoxView.ChildState childState2 = this.createChildState(array[j]);
                    this.stats.add(n + j, childState2);
                    layoutQueue.addTask(childState2);
                }
            }
            layoutQueue.addTask(this.flushTask);
        }
    }
    
    protected void loadChildren(ViewFactory viewFactory) {
        Element element = this.getElement();
        int elementCount = element.getElementCount();
        if (elementCount > 0) {
            View[] array = new View[elementCount];
            for (int i = 0; i < elementCount; ++i) {
                array[i] = viewFactory.create(element.getElement(i));
            }
            this.replace(0, 0, array);
        }
    }
    
    protected synchronized int getViewIndexAtPosition(int n, Position.Bias bias) {
        n = ((bias == Position.Bias.Backward) ? Math.max(0, n - 1) : n);
        return this.getElement().getElementIndex(n);
    }
    
    @Override
    protected void updateLayout(DocumentEvent.ElementChange elementChange, DocumentEvent documentEvent, Shape shape) {
        if (elementChange != null) {
            this.locator.childChanged(this.getChildState(Math.max(elementChange.getIndex() - 1, 0)));
        }
    }
    
    @Override
    public void setParent(View parent) {
        super.setParent(parent);
        if (parent != null && this.getViewCount() == 0) {
            this.loadChildren(this.getViewFactory());
        }
    }
    
    @Override
    public synchronized void preferenceChanged(View view, boolean b, boolean b2) {
        if (view == null) {
            this.getParent().preferenceChanged(this, b, b2);
        }
        else {
            if (this.changing != null && this.changing.getChildView() == view) {
                this.changing.preferenceChanged(b, b2);
                return;
            }
            AsyncBoxView.ChildState childState = this.getChildState(this.getViewIndex(view.getStartOffset(), Position.Bias.Forward));
            childState.preferenceChanged(b, b2);
            LayoutQueue layoutQueue = this.getLayoutQueue();
            layoutQueue.addTask(childState);
            layoutQueue.addTask(this.flushTask);
        }
    }
    
    @Override
    public void setSize(float n, float n2) {
        this.setSpanOnAxis(0, n);
        this.setSpanOnAxis(1, n2);
    }
    
    float getSpanOnAxis(int n) {
        if (n == this.getMajorAxis()) {
            return this.majorSpan;
        }
        return this.minorSpan;
    }
    
    void setSpanOnAxis(int n, float n2) {
        float insetSpan = this.getInsetSpan(n);
        if (n == this.getMinorAxis()) {
            float minorSpan = n2 - insetSpan;
            if (minorSpan != this.minorSpan) {
                this.minorSpan = minorSpan;
                int viewCount = this.getViewCount();
                if (viewCount != 0) {
                    LayoutQueue layoutQueue = this.getLayoutQueue();
                    for (int i = 0; i < viewCount; ++i) {
                        AsyncBoxView.ChildState childState = this.getChildState(i);
                        childState.childSizeValid = false;
                        layoutQueue.addTask(childState);
                    }
                    layoutQueue.addTask(this.flushTask);
                }
            }
        }
        else if (this.estimatedMajorSpan) {
            this.majorSpan = n2 - insetSpan;
        }
    }
    
    @Override
    public void paint(Graphics graphics, Shape allocation) {
        synchronized (this.locator) {
            this.locator.setAllocation(allocation);
            this.locator.paintChildren(graphics);
        }
    }
    
    @Override
    public float getPreferredSpan(int n) {
        float insetSpan = this.getInsetSpan(n);
        if (n == this.axis) {
            return this.majorSpan + insetSpan;
        }
        if (this.prefRequest != null) {
            return this.prefRequest.getChildView().getPreferredSpan(n) + insetSpan;
        }
        return insetSpan + 30.0f;
    }
    
    @Override
    public float getMinimumSpan(int n) {
        if (n == this.axis) {
            return this.getPreferredSpan(n);
        }
        if (this.minRequest != null) {
            return this.minRequest.getChildView().getMinimumSpan(n);
        }
        if (n == 0) {
            return this.getLeftInset() + this.getRightInset() + 5.0f;
        }
        return this.getTopInset() + this.getBottomInset() + 5.0f;
    }
    
    @Override
    public float getMaximumSpan(int n) {
        if (n == this.axis) {
            return this.getPreferredSpan(n);
        }
        return 2.1474836E9f;
    }
    
    @Override
    public int getViewCount() {
        synchronized (this.stats) {
            return this.stats.size();
        }
    }
    
    @Override
    public View getView(int n) {
        AsyncBoxView.ChildState childState = this.getChildState(n);
        if (childState != null) {
            return childState.getChildView();
        }
        return null;
    }
    
    @Override
    public Shape getChildAllocation(int n, Shape shape) {
        return this.locator.getChildAllocation(n, shape);
    }
    
    @Override
    public int getViewIndex(int n, Position.Bias bias) {
        return this.getViewIndexAtPosition(n, bias);
    }
    
    @Override
    public Shape modelToView(int n, Shape shape, Position.Bias bias) throws BadLocationException {
        int viewIndex = this.getViewIndex(n, bias);
        Shape childAllocation = this.locator.getChildAllocation(viewIndex, shape);
        AsyncBoxView.ChildState childState = this.getChildState(viewIndex);
        synchronized (childState) {
            return childState.getChildView().modelToView(n, childAllocation, bias);
        }
    }
    
    @Override
    public int viewToModel(float n, float n2, Shape shape, Position.Bias[] array) {
        int viewIndexAtPoint;
        Shape childAllocation;
        synchronized (this.locator) {
            viewIndexAtPoint = this.locator.getViewIndexAtPoint(n, n2, shape);
            childAllocation = this.locator.getChildAllocation(viewIndexAtPoint, shape);
        }
        AsyncBoxView.ChildState childState = this.getChildState(viewIndexAtPoint);
        int viewToModel;
        synchronized (childState) {
            viewToModel = childState.getChildView().viewToModel(n, n2, childAllocation, array);
        }
        return viewToModel;
    }
    
    @Override
    public int getNextVisualPositionFrom(int n, Position.Bias bias, Shape shape, int n2, Position.Bias[] array) throws BadLocationException {
        if (n < -1) {
            throw new BadLocationException("invalid position", n);
        }
        return Utilities.getNextVisualPositionFrom(this, n, bias, shape, n2, array);
    }
    
    public class ChildLocator
    {
        protected AsyncBoxView.ChildState lastValidOffset;
        protected Rectangle lastAlloc;
        protected Rectangle childAlloc;
        
        public ChildLocator() {
            this.lastAlloc = new Rectangle();
            this.childAlloc = new Rectangle();
        }
        
        public synchronized void childChanged(AsyncBoxView.ChildState childState) {
            if (this.lastValidOffset == null) {
                this.lastValidOffset = childState;
            }
            else if (childState.getChildView().getStartOffset() < this.lastValidOffset.getChildView().getStartOffset()) {
                this.lastValidOffset = childState;
            }
        }
        
        public synchronized void paintChildren(Graphics graphics) {
            Rectangle clipBounds = graphics.getClipBounds();
            int i = this.getViewIndexAtVisualOffset((AsyncBoxView.this.axis == 0) ? ((float)(clipBounds.x - this.lastAlloc.x)) : ((float)(clipBounds.y - this.lastAlloc.y)));
            int viewCount = AsyncBoxView.this.getViewCount();
            float majorOffset = AsyncBoxView.this.getChildState(i).getMajorOffset();
            while (i < viewCount) {
                AsyncBoxView.ChildState childState = AsyncBoxView.this.getChildState(i);
                childState.setMajorOffset(majorOffset);
                Shape childAllocation = this.getChildAllocation(i);
                if (this.intersectsClip(childAllocation, clipBounds)) {
                    synchronized (childState) {
                        childState.getChildView().paint(graphics, childAllocation);
                    }
                    majorOffset += childState.getMajorSpan();
                    ++i;
                }
                else {
                    break;
                }
            }
        }
        
        public synchronized Shape getChildAllocation(int n, Shape allocation) {
            if (allocation == null) {
                return null;
            }
            this.setAllocation(allocation);
            AsyncBoxView.ChildState childState = AsyncBoxView.this.getChildState(n);
            if (this.lastValidOffset == null) {
                this.lastValidOffset = AsyncBoxView.this.getChildState(0);
            }
            if (childState.getChildView().getStartOffset() > this.lastValidOffset.getChildView().getStartOffset()) {
                this.updateChildOffsetsToIndex(n);
            }
            return this.getChildAllocation(n);
        }
        
        public int getViewIndexAtPoint(float n, float n2, Shape allocation) {
            this.setAllocation(allocation);
            return this.getViewIndexAtVisualOffset((AsyncBoxView.this.axis == 0) ? (n - (float)this.lastAlloc.x) : (n2 - (float)this.lastAlloc.y));
        }
        
        protected Shape getChildAllocation(int n) {
            AsyncBoxView.ChildState childState = AsyncBoxView.this.getChildState(n);
            if (!childState.isLayoutValid()) {
                childState.run();
            }
            if (AsyncBoxView.this.axis == 0) {
                this.childAlloc.x = this.lastAlloc.x + (int)childState.getMajorOffset();
                this.childAlloc.y = this.lastAlloc.y + (int)childState.getMinorOffset();
                this.childAlloc.width = (int)childState.getMajorSpan();
                this.childAlloc.height = (int)childState.getMinorSpan();
            }
            else {
                this.childAlloc.y = this.lastAlloc.y + (int)childState.getMajorOffset();
                this.childAlloc.x = this.lastAlloc.x + (int)childState.getMinorOffset();
                this.childAlloc.height = (int)childState.getMajorSpan();
                this.childAlloc.width = (int)childState.getMinorSpan();
            }
            Rectangle childAlloc = this.childAlloc;
            childAlloc.x += (int)AsyncBoxView.this.getLeftInset();
            Rectangle childAlloc2 = this.childAlloc;
            childAlloc2.y += (int)AsyncBoxView.this.getRightInset();
            return this.childAlloc;
        }
        
        protected void setAllocation(Shape shape) {
            if ((shape instanceof Rectangle)) {
                this.lastAlloc.setBounds((Rectangle)shape);
            }
            else {
                this.lastAlloc.setBounds(shape.getBounds());
            }
            AsyncBoxView.this.setSize((float)this.lastAlloc.width, (float)this.lastAlloc.height);
        }
        
        protected int getViewIndexAtVisualOffset(float n) {
            int viewCount = AsyncBoxView.this.getViewCount();
            if (viewCount > 0) {
                int n2 = this.lastValidOffset != null;
                if (this.lastValidOffset == null) {
                    this.lastValidOffset = AsyncBoxView.this.getChildState(0);
                }
                if (n > AsyncBoxView.this.majorSpan) {
                    if (n2 == 0) {
                        return 0;
                    }
                    return AsyncBoxView.this.getViewIndex(this.lastValidOffset.getChildView().getStartOffset(), Position.Bias.Forward);
                }
                else {
                    if (n > this.lastValidOffset.getMajorOffset()) {
                        return this.updateChildOffsets(n);
                    }
                    float n3 = 0.0f;
                    int i = 0;
                    while (i < viewCount) {
                        float n4 = n3 + AsyncBoxView.this.getChildState(i).getMajorSpan();
                        if (n < n4) {
                            return i;
                        }
                        n3 = n4;
                        ++i;
                    }
                }
            }
            return viewCount - 1;
        }
        
        int updateChildOffsets(float n) {
            int viewCount = AsyncBoxView.this.getViewCount();
            int n2 = viewCount - 1;
            int i = AsyncBoxView.this.getViewIndex(this.lastValidOffset.getChildView().getStartOffset(), Position.Bias.Forward);
            float majorOffset = this.lastValidOffset.getMajorOffset();
            while (i < viewCount) {
                AsyncBoxView.ChildState childState = AsyncBoxView.this.getChildState(i);
                childState.setMajorOffset(majorOffset);
                majorOffset += childState.getMajorSpan();
                if (n < majorOffset) {
                    n2 = i;
                    this.lastValidOffset = childState;
                    break;
                }
                ++i;
            }
            return n2;
        }
        
        void updateChildOffsetsToIndex(int n) {
            int i;
            float majorOffset;
            AsyncBoxView.ChildState childState;
            for (i = AsyncBoxView.this.getViewIndex(this.lastValidOffset.getChildView().getStartOffset(), Position.Bias.Forward), majorOffset = this.lastValidOffset.getMajorOffset(), ; i <= n; majorOffset += childState.getMajorSpan(), ++i) {
                childState = AsyncBoxView.this.getChildState(i);
                childState.setMajorOffset(majorOffset);
            }
        }
        
        boolean intersectsClip(Shape shape, Rectangle r) {
            Rectangle r2 = (shape instanceof Rectangle) ? ((Rectangle)shape) : shape.getBounds();
            return r2.intersects(r) && this.lastAlloc.intersects(r2);
        }
    }
    
    public class ChildState implements Runnable
    {
        private float min;
        private float pref;
        private float max;
        private boolean minorValid;
        private float span;
        private float offset;
        private boolean majorValid;
        private View child;
        private boolean childSizeValid;
        
        public ChildState(View child) {
            this.child = child;
            this.minorValid = false;
            this.majorValid = false;
            this.childSizeValid = false;
            this.child.setParent(AsyncBoxView.this);
        }
        
        public View getChildView() {
            return this.child;
        }
        
        @Override
        public void run() {
            AbstractDocument abstractDocument = (AbstractDocument)AsyncBoxView.this.getDocument();
            try {
                abstractDocument.readLock();
                if (this.minorValid && this.majorValid && this.childSizeValid) {
                    return;
                }
                if (this.child.getParent() == AsyncBoxView.this) {
                    synchronized (AsyncBoxView.this) {
                        AsyncBoxView.this.changing = this;
                    }
                    this.updateChild();
                    synchronized (AsyncBoxView.this) {
                        AsyncBoxView.this.changing = null;
                    }
                    this.updateChild();
                }
            }
            finally {
                abstractDocument.readUnlock();
            }
        }
        
        void updateChild() {
            boolean n = false;
            synchronized (this) {
                if (!this.minorValid) {
                    int minorAxis = AsyncBoxView.this.getMinorAxis();
                    this.min = this.child.getMinimumSpan(minorAxis);
                    this.pref = this.child.getPreferredSpan(minorAxis);
                    this.max = this.child.getMaximumSpan(minorAxis);
                    this.minorValid = true;
                    n = true;
                }
            }
            if (!(n == 0)) {
                AsyncBoxView.this.minorRequirementChange(this);
            }
            boolean n2 = false;
            float n3 = 0.0f;
            synchronized (this) {
                if (!this.majorValid) {
                    float span = this.span;
                    this.span = this.child.getPreferredSpan(AsyncBoxView.this.axis);
                    n3 = this.span - span;
                    this.majorValid = true;
                    n2 = true;
                }
            }
            if (!(n2 == 0)) {
                AsyncBoxView.this.majorRequirementChange(this, n3);
                AsyncBoxView.this.locator.childChanged(this);
            }
            synchronized (this) {
                if (!this.childSizeValid) {
                    float n4;
                    float n5;
                    if (AsyncBoxView.this.axis == 0) {
                        n4 = this.span;
                        n5 = this.getMinorSpan();
                    }
                    else {
                        n4 = this.getMinorSpan();
                        n5 = this.span;
                    }
                    this.childSizeValid = true;
                    this.child.setSize(n4, n5);
                }
            }
        }
        
        public float getMinorSpan() {
            if (this.max < AsyncBoxView.this.minorSpan) {
                return this.max;
            }
            return Math.max(this.min, AsyncBoxView.this.minorSpan);
        }
        
        public float getMinorOffset() {
            if (this.max < AsyncBoxView.this.minorSpan) {
                return (AsyncBoxView.this.minorSpan - this.max) * this.child.getAlignment(AsyncBoxView.this.getMinorAxis());
            }
            return 0.0f;
        }
        
        public float getMajorSpan() {
            return this.span;
        }
        
        public float getMajorOffset() {
            return this.offset;
        }
        
        public void setMajorOffset(float offset) {
            this.offset = offset;
        }
        
        public void preferenceChanged(boolean b, boolean b2) {
            if (AsyncBoxView.this.axis == 0) {
                if (b) {
                    this.majorValid = false;
                }
                if (b2) {
                    this.minorValid = false;
                }
            }
            else {
                if (b) {
                    this.minorValid = false;
                }
                if (b2) {
                    this.majorValid = false;
                }
            }
            this.childSizeValid = false;
        }
        
        public boolean isLayoutValid() {
            return this.minorValid && this.majorValid && this.childSizeValid;
        }
    }
    
    class FlushTask implements Runnable
    {
        @Override
        public void run() {
            AsyncBoxView.this.flushRequirementChanges();
        }
    }
}
